/* dhcpc.c
 *
 * udhcp DHCP client
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */

#include <sys/time.h>
#include <sys/file.h>
#include <sys/sysinfo.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "common.h"
#include "dhcpd.h"
#include "dhcpc.h"
#include "options.h"
#include "clientpacket.h"
#include "clientsocket.h"
#include "script.h"
#include "socket.h"
#include "signalpipe.h"
#ifdef UDHCP_ZEROCONF
#include "zeroconf.h"
#endif
 
#include "csman.h"
#include "unilog.h"
#include "csid/csid_dhcp.h"

static int state;
static unsigned long requested_ip; /* = 0 */
static unsigned long server_addr;
static unsigned long timeout;
static int packet_num; /* = 0 */
static int fd = -1;

#define LISTEN_NONE 0
#define LISTEN_KERNEL 1
#define LISTEN_RAW 2
#define LISTEN_DEMAND 3
static int listen_mode;

#define DEMAND_FILE		"/var/run/udhcpc.demand"

/* Warning: 2k038 bug :P */

#define TIMEOUT_INFINITE 0x7fffffff

struct client_config_t client_config = {
	/* Default options. */
	.abort_if_no_lease = 0,
	.foreground = 0,
	.quit_after_lease = 0,
	.background_if_no_lease = 0,
	.interface = "eth0",
	.wanIndex = NULL,
#if defined(_ENABLE_VOIPWAN)
	.voipwanIndex = NULL,
#endif
	.pidfile = NULL,
	.script = DEFAULT_SCRIPT,
	.clientid = NULL,
	.vendorclass = NULL,
	.hostname = NULL,
	.fqdn = NULL,
	.ifindex = 0,
	.demand = 0,
    .lanif = "br0",
	.retries = 3,
	.timeout = 3,
	.arp = "\0\0\0\0\0\0",		/* appease gcc-3.0 */
};

#ifndef IN_BUSYBOX
static void ATTRIBUTE_NORETURN show_usage(void)
{
	printf(
"Usage: udhcpc [OPTIONS]\n\n"
"  -c, --clientid=CLIENTID         Set client identifier - type is first char\n"
"  -C, --clientid-none             Suppress default client identifier\n"
"  -V, --vendorclass=CLASSID       Set vendor class identifier\n"
"  -H, --hostname=HOSTNAME         Client hostname\n"
"  -h                              Alias for -H\n"
"  -F, --fqdn=FQDN                 Client fully qualified domain name\n"
"  -f, --foreground                Do not fork after getting lease\n"
"  -b, --background                Fork to background if lease cannot be\n"
"                                  immediately negotiated.\n"
"  -i, --interface=INTERFACE       Interface to use (default: eth0)\n"
"  -I, --wanindex			       MultiWAN index\n"
#if defined(_ENABLE_VOIPWAN)
"  -W, --voipwan				   VoIP WAN index\n"
#endif
"  -z, --zeroconf                  With Zeroconf address attibution support\n"
"  -n, --now                       Exit with failure if lease cannot be\n"
"                                  immediately negotiated.\n"
"  -p, --pidfile=file              Store process ID of daemon in file\n"
"  -q, --quit                      Quit after obtaining lease\n"
"  -r, --request=IP                IP address to request (default: none)\n"
"  -s, --script=file               Run file at dhcp events (default:\n"
"                                  " DEFAULT_SCRIPT ")\n"
"  -T, --timeout=seconds           Try to get the lease for the amount of\n"
"                                  seconds (default: 3)\n"
"  -t, --retries=NUM               Send up to NUM request packets\n"
"  -d, --demand					  Connect on Demand\n"					  	
"  -l, --lanif					  lan interface name\n"
"  -v, --version                   Display version\n"
	);
	exit(0);
}
#else
#define show_usage bb_show_usage
extern void show_usage(void) ATTRIBUTE_NORETURN;
#endif


/* just a little helper */
static void change_mode(int new_mode)
{
	DEBUG(LOG_INFO, "entering %s listen mode",
		new_mode ? (new_mode == 1 ? "kernel" : "raw") : "none");
	if (fd >= 0) close(fd);
	fd = -1;
	listen_mode = new_mode;
}


/* perform a renew */
static void perform_renew(void)
{
	LOG(LOG_INFO, "Performing a DHCP renew");
	switch (state) {
	case BOUND:
		change_mode(LISTEN_KERNEL);
	case RENEWING:
	case REBINDING:
		state = RENEW_REQUESTED;
		break;
	case RENEW_REQUESTED: /* impatient are we? fine, square 1 */
		run_script(NULL, "deconfig");
#ifdef UDHCP_ZEROCONF
		zeroconf_event(ZEROCONF_EVENT_DHCPOUT);
#endif
	case REQUESTING:
	case RELEASED:
		change_mode(LISTEN_RAW);
		state = INIT_SELECTING;
		break;
	case INIT_SELECTING:
		break;
	}

	/* start things over */
	packet_num = 0;

	/* Kill any timeouts because the user wants this to hurry along */
	timeout = 0;
}


/* perform a release */
static void perform_release(void)
{
	char buffer[16];
	struct in_addr temp_addr;
	/* send release packet */
	if (state == BOUND || state == RENEWING || state == REBINDING) {
		temp_addr.s_addr = server_addr;
		sprintf(buffer, "%s", inet_ntoa(temp_addr));
		temp_addr.s_addr = requested_ip;
		LOG(LOG_INFO, "Unicasting a release of %s to %s",
				inet_ntoa(temp_addr), buffer);
#ifdef UDHCP_ZEROCONF
		zeroconf_event(ZEROCONF_EVENT_DHCPOUT);
#endif
		send_release(server_addr, requested_ip); /* unicast */
		run_script(NULL, "deconfig");
	}
	LOG(LOG_INFO, "Entering released state");

	change_mode(LISTEN_NONE);
	state = RELEASED;
	timeout = TIMEOUT_INFINITE;
}

static void client_background(void)
{
	background(client_config.pidfile);
	client_config.foreground = 1; /* Do not fork again. */
	client_config.background_if_no_lease = 0;
}


#ifdef COMBINED_BINARY
int udhcpc_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
	uint8_t *temp, *message;
	unsigned long t1 = 0, t2 = 0, xid = 0;
	unsigned long start = 0, lease,lease_csid;
	fd_set rfds;
	int retval;
	struct timeval tv;
	int c, len;
	struct dhcpMessage packet;
	struct in_addr temp_addr;
	unsigned long now;
	int max_fd;
	int sig;
#ifdef UDHCP_ZEROCONF
 	int with_zeroconf = 0;
#endif
	int no_clientid = 0;
	u_int32_t lan_ip;
    u_int32_t lan_mask;
    char buf[200];
    struct sysinfo info;
    int csman_fd;

	static const struct option arg_options[] = {
		{"clientid",	required_argument,	0, 'c'},
		{"clientid-none", no_argument,		0, 'C'},
		{"vendorclass",	required_argument,	0, 'V'},
		{"foreground",	no_argument,		0, 'f'},
		{"background",	no_argument,		0, 'b'},
		{"hostname",	required_argument,	0, 'H'},
		{"hostname",	required_argument,	0, 'h'},
		{"fqdn",	required_argument,	0, 'F'},
		{"interface",	required_argument,	0, 'i'},
		{"wanindex",	required_argument,	0, 'I'},		
#if defined(_ENABLE_VOIPWAN)
		{"voipwanindex",	required_argument,	0, 'W'},
#endif		
		{"zeroconf",	no_argument,		0, 'z'},
		{"now",		no_argument,		0, 'n'},
		{"pidfile",	required_argument,	0, 'p'},
		{"quit",	no_argument,		0, 'q'},
		{"request",	required_argument,	0, 'r'},
		{"script",	required_argument,	0, 's'},
		{"timeout",	required_argument,	0, 'T'},
		{"version",	no_argument,		0, 'v'},
        {"demand", no_argument,			0, 'd'},
        {"lanif", required_argument, 0, 'l'},
		{"retries",	required_argument,	0, 't'},		
		{0, 0, 0, 0}
	};

	/* get options */
	while (1) {
		int option_index = 0;

#if defined(_ENABLE_VOIPWAN)
		c = getopt_long(argc, argv, "c:CV:fbH:h:F:i:I:W:znp:qr:s:T:t:vdl:", arg_options, &option_index);
#else
		c = getopt_long(argc, argv, "c:CV:fbH:h:F:i:I:znp:qr:s:T:t:vdl:", arg_options, &option_index);
#endif

		if (c == -1) break;

		switch (c) {
		case 'c':
			if (no_clientid) show_usage();
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			free(client_config.clientid);
			client_config.clientid = xmalloc(len + 2);
			client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
			client_config.clientid[OPT_LEN] = len;
			client_config.clientid[OPT_DATA] = '\0';
			strncpy((char*)client_config.clientid + OPT_DATA, optarg, len);
			break;
		case 'C':
			if (client_config.clientid) show_usage();
			no_clientid = 1;
			break;
		case 'V':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			free(client_config.vendorclass);
			client_config.vendorclass = xmalloc(len + 2);
			client_config.vendorclass[OPT_CODE] = DHCP_VENDOR;
			client_config.vendorclass[OPT_LEN] = len;
			strncpy((char*)client_config.vendorclass + OPT_DATA, optarg, len);
			break;
		case 'f':
			client_config.foreground = 1;
			break;
		case 'b':
			client_config.background_if_no_lease = 1;
			break;
		case 'h':
		case 'H':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			free(client_config.hostname);
			client_config.hostname = xmalloc(len + 2);
			client_config.hostname[OPT_CODE] = DHCP_HOST_NAME;
			client_config.hostname[OPT_LEN] = len;
			strncpy((char*)client_config.hostname + 2, optarg, len);
			break;
		case 'F':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			free(client_config.fqdn);
			client_config.fqdn = xmalloc(len + 5);
			client_config.fqdn[OPT_CODE] = DHCP_FQDN;
			client_config.fqdn[OPT_LEN] = len + 3;
			/* Flags: 0000NEOS
			S: 1 => Client requests Server to update A RR in DNS as well as PTR
			O: 1 => Server indicates to client that DNS has been updated regardless
			E: 1 => Name data is DNS format, i.e. <4>host<6>domain<4>com<0> not "host.domain.com"
			N: 1 => Client requests Server to not update DNS
			*/
			client_config.fqdn[OPT_LEN + 1] = 0x1;
			client_config.fqdn[OPT_LEN + 2] = 0;
			client_config.fqdn[OPT_LEN + 3] = 0;
			strncpy((char*)client_config.fqdn + 5, optarg, len);
			break;
		case 'i':
			client_config.interface =  optarg;
			break;
		case 'I':
			client_config.wanIndex = optarg;
			break;
#if defined(_ENABLE_VOIPWAN)
		case 'W':
			client_config.voipwanIndex = optarg;
			break;
#endif
		case 'z':
#ifdef UDHCP_ZEROCONF
			with_zeroconf = 1;
#else
			printf("Zeroconf support not included in this build.\n");
#endif
			break;
		case 'n':
			client_config.abort_if_no_lease = 1;
			break;
		case 'p':
			client_config.pidfile = optarg;
			break;
		case 'q':
			client_config.quit_after_lease = 1;
			break;
		case 'r':
			requested_ip = inet_addr(optarg);
			break;
		case 's':
			client_config.script = optarg;
			break;
		case 'T':
			client_config.timeout = atoi(optarg);
			break;
		case 't':
			client_config.retries = atoi(optarg);
			break;
		case 'v':
			printf("udhcpcd, version %s\n\n", VERSION);
			return 0;
			break;
        case 'd':
        	client_config.demand=1;
        	break;    
        case 'l':        	
        	client_config.lanif=optarg;        	
        	break;    
		default:
			show_usage();
		}
	}

	/* Start the log, sanitize fd's, and write a pid file */
	start_log_and_pid("udhcpc", client_config.pidfile);

	if (read_interface(client_config.interface, &client_config.ifindex,
			   NULL, client_config.arp) < 0)
		return 1;

#ifdef UDHCP_ZEROCONF
	zeroconf_init(with_zeroconf, client_config.arp, client_config.interface, 
	client_config.ifindex);
#endif

	/* if not set, and not suppressed, setup the default client ID */
	if (!client_config.clientid && !no_clientid) {
		client_config.clientid = xmalloc(6 + 3);
		client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
		client_config.clientid[OPT_LEN] = 7;
		client_config.clientid[OPT_DATA] = 1;
		memcpy(client_config.clientid + 3, client_config.arp, 6);
	}

	if (!client_config.vendorclass) {
		client_config.vendorclass = xmalloc(sizeof("udhcp "VERSION) + 2);
		client_config.vendorclass[OPT_CODE] = DHCP_VENDOR;
		client_config.vendorclass[OPT_LEN] = sizeof("udhcp "VERSION) - 1;
		client_config.vendorclass[OPT_DATA] = 1;
		memcpy(&client_config.vendorclass[OPT_DATA],
			"udhcp "VERSION, sizeof("udhcp "VERSION) - 1);
	}

	/* setup the signal pipe */
	udhcp_sp_setup();

	state = INIT_SELECTING;
	run_script(NULL, "deconfig");

#ifdef UDHCP_ZEROCONF
	zeroconf_event(ZEROCONF_EVENT_DHCPOUT);
#endif

	if( client_config.demand){
		char rgch[32];
		unsigned char zc_addr[4]={169,254,91,77};
		
		sprintf(rgch, "%d.%d.%d.%d", zc_addr[0], zc_addr[1], zc_addr[2], zc_addr[3]);
		
		run_script_zeroconf(rgch, "255.255.0.0", "bound");      
		change_mode(LISTEN_DEMAND);
		if(client_config.wanIndex == NULL){
			sprintf(buf, "echo 'demand' > %s", DEMAND_FILE);
		}else{
			sprintf(buf, "echo 'demand' > /var/run/udhcpc%d.demand", atoi(client_config.wanIndex)+1);
		}
		system(buf);
		client_background();
	}else{	
    	change_mode(LISTEN_RAW);
	}

	for (;;) {

     		/* configures timeout for select(), the smaller between DHCP and Zeroconf timeouts */
#ifdef UDHCP_ZEROCONF
     		zeroconf_set_smallest_timeout(&tv, timeout);
#else
		tv.tv_sec = timeout - time(NULL);
		if (tv.tv_sec < 0) {
			tv.tv_sec = 0;
		}
		tv.tv_usec = 0;
#endif

		if (listen_mode != LISTEN_NONE && fd < 0) {
			if (listen_mode == LISTEN_KERNEL)
				fd = listen_socket(INADDR_ANY, CLIENT_PORT, client_config.interface);
			else if ( listen_mode == LISTEN_RAW )
                fd = raw_socket(client_config.ifindex);
            else if( listen_mode == LISTEN_DEMAND)
            	fd=raw_demand_socket(client_config.lanif, &lan_ip, &lan_mask);
           
			if (fd < 0) {
				LOG(LOG_ERR, "FATAL: couldn't listen on socket, %m");
				return 0;
			}
		}

#ifdef UDHCP_ZEROCONF
		if (zeroconf_fd > -1) {
			max_fd = udhcp_sp_fd_set2(&rfds, fd, zeroconf_fd);
		} else {
			max_fd = udhcp_sp_fd_set(&rfds, fd);
		}
#else
		max_fd = udhcp_sp_fd_set(&rfds, fd);
#endif
		if (tv.tv_sec >= 0 || tv.tv_usec > 0) {
			DEBUG(LOG_INFO, "Waiting on select...");			
			if( listen_mode == LISTEN_DEMAND ){
            	retval = select(max_fd + 1, &rfds, NULL, NULL, NULL);
        	}else{
             	retval = select(max_fd + 1, &rfds, NULL, NULL, &tv);
	        }			
		} else retval = 0; /* If we already timed out, fall through */

		now = time(NULL);
		if (retval == 0) {
			/* timeout, must be forwarded to Zeroconf and DHCP state machines */

#ifdef UDHCP_ZEROCONF
			/* Zeroconf has its own timeout tests, just tick it */
			zeroconf_event(ZEROCONF_EVENT_TICK);
#endif

			if (timeout > now) {
				/* it is not DHCP's timeout */
				continue;
			}

			DEBUG(LOG_INFO, "DHCP timeout");
			
			/* DHCP timeout dropped to zero */
			switch (state) {
			case INIT_SELECTING:
			
				if( listen_mode == LISTEN_DEMAND){					
					timeout=time(0);            		
            		break;
            	}	
				
				if (packet_num < client_config.retries) {
					if (packet_num == 0)
						xid = random_xid();

					/* send discover packet */
					send_discover(xid, requested_ip); /* broadcast */

					timeout = now + client_config.timeout;
					packet_num++;
				} else {
					run_script(NULL, "leasefail");
					if (client_config.background_if_no_lease) {
						LOG(LOG_INFO, "No lease, forking to background.");
						client_background();
					} else if (client_config.abort_if_no_lease) {
						LOG(LOG_INFO, "No lease, failing.");
						return 1;
					}
					/* wait to try again */
					packet_num = 0;
					timeout = now + 60;
				}
				break;
			case RENEW_REQUESTED:
			case REQUESTING:
				if (packet_num < client_config.retries) {
					/* send request packet */
					if (state == RENEW_REQUESTED)
						send_renew(xid, server_addr, requested_ip); /* unicast */
					else send_selecting(xid, server_addr, requested_ip); /* broadcast */

					timeout = now + ((packet_num == 2) ? 10 : 2);
					packet_num++;
				} else {
					/* timed out, go back to init state */
					if (state == RENEW_REQUESTED) run_script(NULL, "deconfig");
#ifdef UDHCP_ZEROCONF
					zeroconf_event(ZEROCONF_EVENT_DHCPOUT);
#endif
					state = INIT_SELECTING;
					timeout = now;
					packet_num = 0;
					change_mode(LISTEN_RAW);
				}
				break;
			case BOUND:
				/* Lease is starting to run out, time to enter renewing state */
				state = RENEWING;
				change_mode(LISTEN_KERNEL);
				DEBUG(LOG_INFO, "Entering renew state");
				/* fall right through */
			case RENEWING:
				/* Either set a new T1, or enter REBINDING state */
				if ((t2 - t1) <= (lease / 14400 + 1)) {
					/* timed out, enter rebinding state */
					state = REBINDING;
					timeout = now + (t2 - t1);
					DEBUG(LOG_INFO, "Entering rebinding state");
				} else {
					/* send a request packet */
					send_renew(xid, server_addr, requested_ip); /* unicast */

					t1 = (t2 - t1) / 2 + t1;
					timeout = t1 + start;
				}
				break;
			case REBINDING:
				/* Either set a new T2, or enter INIT state */
				if ((lease - t2) <= (lease / 14400 + 1)) {
					/* timed out, enter init state */
					state = INIT_SELECTING;
					LOG(LOG_INFO, "Lease lost, entering init state");
					run_script(NULL, "deconfig");
#ifdef UDHCP_ZEROCONF
					zeroconf_event(ZEROCONF_EVENT_DHCPOUT);
#endif
					timeout = now;
					packet_num = 0;
					change_mode(LISTEN_RAW);
				} else {
					/* send a request packet */
					send_renew(xid, 0, requested_ip); /* broadcast */

					t2 = (lease - t2) / 2 + t2;
					timeout = t2 + start;
				}
				break;
			case RELEASED:
				/* yah, I know, *you* say it would never happen */
				timeout = TIMEOUT_INFINITE;
				break;
			}
#ifdef UDHCP_ZEROCONF
		} else if (retval > 0 && zeroconf_fd > -1 && FD_ISSET(zeroconf_fd, &rfds)) {
			/* zeroconf packet is ready, process */

			DEBUG(LOG_INFO, "Socket data for Zeroconf");

			zeroconf_event(ZEROCONF_EVENT_SOCKETREADY);
#endif
		} else if (retval > 0 && listen_mode != LISTEN_NONE && FD_ISSET(fd, &rfds)) {

			DEBUG(LOG_INFO, "Socket data for DHCP");

			/* a packet is ready, read it */

			 if (listen_mode == LISTEN_KERNEL){
                len = get_packet(&packet, fd);
            }else if( listen_mode == LISTEN_RAW){
                len = get_raw_packet(&packet, fd);
            }else if( listen_mode == LISTEN_DEMAND){
               	retval=check_raw_demand_packet(fd, lan_ip, lan_mask);
	        	if(retval<0){
	        		LOG(LOG_ERR,"FATAL: check_raw_packet_for_demand failed");
	        		return 0; 
	        	}else if( retval>0){ // OK!!	        		
        			timeout=0;
        			change_mode(LISTEN_RAW);
				if(client_config.wanIndex == NULL){
        				sprintf(buf, "echo 'raw' > %s", DEMAND_FILE);
				}else{
					sprintf(buf, "echo 'raw' > /var/run/udhcpc%d.demand", atoi(client_config.wanIndex)+1);
				}
					system(buf);
	        		continue;
	        	}else if( retval==0){ // Failed!!	        		
	        		timeout=time(0);	
	        		continue;
	        	}				
            }	     
			

			if (len == -1 && errno != EINTR) {
				DEBUG(LOG_INFO, "error on read, %m, reopening socket");
				change_mode(listen_mode); /* just close and reopen */
			}
			if (len < 0) continue;

			if (packet.xid != xid) {
				DEBUG(LOG_INFO, "Ignoring XID %lx (our xid is %lx)",
					(unsigned long) packet.xid, xid);
				continue;
			}

			/* Ignore packets that aren't for us */
			if (memcmp(packet.chaddr, client_config.arp, 6)) {
				DEBUG(LOG_INFO, "packet does not have our chaddr -- ignoring");
				continue;
			}

			if ((message = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
				DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
				continue;
			}

			switch (state) {
			case INIT_SELECTING:
				/* Must be a DHCPOFFER to one of our xid's */
				if (*message == DHCPOFFER) {
					if ((temp = get_option(&packet, DHCP_SERVER_ID))) {
						memcpy(&server_addr, temp, 4);
						xid = packet.xid;
						requested_ip = packet.yiaddr;

						/* enter requesting state */
						state = REQUESTING;
						timeout = now;
						packet_num = 0;
					} else {
						DEBUG(LOG_ERR, "No server ID in message");
					}
				}
				break;
			case RENEW_REQUESTED:
			case REQUESTING:
			case RENEWING:
			case REBINDING:
				if (*message == DHCPACK) {
					if (!(temp = get_option(&packet, DHCP_LEASE_TIME))) {
						LOG(LOG_ERR, "No lease time with ACK, using 1 hour lease");
						lease = 60 * 60;
					} else {
						memcpy(&lease, temp, 4);
						lease = ntohl(lease);
					}

					sysinfo(&info);
                    csman_fd=open_csman(NULL,0);
                    lease_csid=info.uptime+lease;

#if defined(_SMS_REMOTE_MANAGEMENT)
                    write_csman(csman_fd, CSID_S_DHCP_CLI_UP_TIME, &info.uptime, sizeof(info.uptime),0);
#endif
    
					if(client_config.wanIndex == NULL){
#if defined(_ENABLE_VOIPWAN)
						if(client_config.voipwanIndex == NULL){
#endif
                    		write_csman(csman_fd, CSID_S_DHCP_CLI_LEASE_TIME, (const void *)&lease_csid, sizeof(unsigned long), 0);
#if defined(_ENABLE_VOIPWAN)
						}else{
							write_csman(csman_fd, CSID_S_DHCP_VOIPWAN_LEASE_TIME+atoi(client_config.voipwanIndex), (const void *)&lease_csid, sizeof(unsigned long), 0);
						}
#endif
					}else{
                    	write_csman(csman_fd, CSID_S_DHCP_MULTIWAN_LEASE_TIME+atoi(client_config.wanIndex), (const void *)&lease_csid, sizeof(unsigned long), 0);
					}
					close_csman(csman_fd);	
					LPRINTF(PRI_D1, " lease_csid=%ld uptime=%ld lease=%ld\n", lease_csid, info.uptime, lease);
                   
					/* enter bound state */
					t1 = lease / 2;

					/* little fixed point for n * .875 */
					t2 = (lease * 0x7) >> 3;
					temp_addr.s_addr = packet.yiaddr;
					LOG(LOG_INFO, "Lease of %s obtained, lease time %ld",
						inet_ntoa(temp_addr), lease);
#ifdef	_CDS530AM_CONNECT
					system("killall amit_avs");
					int csmfd_connect = open_csman(NULL,0);
					int confirm_got_ip=1;
					write_csman(csmfd_connect,CSID_S_DHCP_AUTO_MODE_GET_IP,&confirm_got_ip,sizeof(int),0);
					close_csman(csmfd_connect);
					system("amit_avs &");
#endif
					start = now;
					timeout = t1 + start;
					requested_ip = packet.yiaddr;
#ifdef UDHCP_ZEROCONF
					zeroconf_event(ZEROCONF_EVENT_DHCPIN);
#endif
					run_script(&packet,
						   ((state == RENEWING || state == REBINDING || state == RENEW_REQUESTED) ? "renew" : "bound"));

					state = BOUND;
					change_mode(LISTEN_NONE);
					if (client_config.quit_after_lease)
						return 0;
					if (!client_config.foreground)
						client_background();

				} else if (*message == DHCPNAK) {
					/* return to init state */
					LOG(LOG_INFO, "Received DHCP NAK");
					run_script(&packet, "nak");
					if (state != REQUESTING) {
						run_script(NULL, "deconfig");
#ifdef UDHCP_ZEROCONF
						zeroconf_event(ZEROCONF_EVENT_DHCPOUT);
#endif
					}
					state = INIT_SELECTING;
					timeout = now;
					requested_ip = 0;
					packet_num = 0;
					change_mode(LISTEN_RAW);
					sleep(3); /* avoid excessive network traffic */
				}
				break;
			/* case BOUND, RELEASED: - ignore all packets */
			}			
		} else if (retval > 0 && (sig = udhcp_sp_read(&rfds))) {
			switch (sig) {
			case SIGUSR1:
				perform_renew();
				break;
			case SIGUSR2:
				perform_release();
				break;
			case SIGTERM:
				LOG(LOG_INFO, "Received SIGTERM");
				return 0;
			}
		} else if (retval == -1 && errno == EINTR) {
			/* a signal was caught */
		} else {
			/* An error occured */
			DEBUG(LOG_ERR, "Error on select");
		}

	}
	return 0;
}
